# Standard Library
from typing import Any, List

# First Party
from minislite.tables import TableManager
from minislite.exceptions import AreYouSure, AlreadyExistsError, RecordNotFoundError


class Objects:
    table_name = None
    unique_together = None

    klass: Any
    manager: Any

    def __get__(self, instance, owner):
        self.table_name = owner.get_table_name()
        self.unique_together = owner.get_unique_together()
        self.klass = owner
        self.manager = TableManager(table_name=self.table_name)

        return self

    def get(self, **kwargs) -> Any:
        db_object = self.manager.select(**kwargs).fetchone()
        if not db_object:
            raise RecordNotFoundError("This row not found.")

        dict_object = dict(db_object)
        return self.klass(**dict_object)

    def filter(self, **kwargs) -> List[Any]:
        db_objects = self.manager.select(**kwargs).fetchall()

        object_list = [self.klass(**dict(i)) for i in db_objects]
        return object_list

    def all(self) -> List[Any]:
        return self.filter()

    def check_unique_together(self, is_creating=False, object_id=None, **kwargs) -> bool:
        if not self.unique_together:
            return True
        fields = {}
        for unique in self.unique_together:
            fields[unique] = kwargs.get(unique, None)

        if is_creating:
            check = self.filter(**fields)
            if check:
                return False
            return True

        db_object = self.filter(**fields)
        if db_object:
            if db_object[0].id != object_id:
                return False
        return True

    def update(self, *args, **kwargs):
        if self.check_unique_together(*args, **kwargs):
            self.manager.update(*args, **kwargs)
        else:
            raise AlreadyExistsError("This object already exists.")

    def create(self, *args, **kwargs):
        if self.check_unique_together(*args, is_creating=True, **kwargs):  # type: ignore
            object_id = self.manager.insert(*args, **kwargs)
            return self.get(id=object_id)

        raise AlreadyExistsError("This object already exists.")

    def delete(self, i_am_sure=None, object_id=None) -> None:
        if object_id is None and i_am_sure is None:
            raise AreYouSure("Cleaning all data in tables. Are you sure? Use 'i_am_sure=True'")

        self.manager.delete(object_id)



# Standard Library
from typing import List

# First Party
from minislite.fields import DatabaseField
from minislite.tables import TableManager
from minislite.objects import Objects


class MiniSLiteModel:
    id = DatabaseField(field_type=int, auto_increment=True)

    table_name = None
    unique_together: List[str] = []
    objects = Objects()

    def __init__(self, **kwargs):
        self.reload_attributes(**kwargs)
        self.manager = TableManager(table_name=self.get_table_name())

    def reload_attributes(self, **kwargs) -> None:
        field_list = self.get_fields()
        for field in field_list:
            exists_attr = getattr(self, field, None)
            if isinstance(exists_attr, DatabaseField):
                exists_attr = exists_attr.default
            value = kwargs.get(field, exists_attr)
            setattr(self, field, value)

        setattr(self, "id", kwargs.get("id", None))

    def save(self):
        field_list = self.get_fields()
        fields = {}
        for field in field_list:
            fields[field] = getattr(self, field, None)

        if not self.id:
            this_object = self.objects.create(**fields)
            self.id = this_object.id
        else:
            self.objects.update(object_id=self.id, **fields)

    def delete(self):
        if self.id:
            self.objects.delete(object_id=self.id)
            self.id = None  # type: ignore

    @classmethod
    def get_unique_together(cls) -> List[str]:
        return cls.unique_together

    @classmethod
    def get_table_name(cls) -> str:
        table_name = cls.table_name
        if table_name is None:
            table_name = cls.__name__.lower()
        return table_name

    @classmethod
    def get_fields(cls) -> List[str]:
        field_list = []
        for (field, value) in cls.__dict__.items():
            if isinstance(value, DatabaseField):
                field_list.append(field)

        return field_list



# Standard Library
import os
import sqlite3

# First Party
from minislite.exceptions import DatabaseNotFound


class TableManager:
    def __init__(self, table_name) -> None:
        db_file = os.environ.get("MINILITE_DB_PATH", None)
        if db_file is None:
            raise DatabaseNotFound("Database not defined.")

        self.table_name = table_name
        self.connection = sqlite3.connect(db_file)
        self.connection.row_factory = sqlite3.Row

        self.cursor = self.connection.cursor()

    def select(self, **kwargs) -> sqlite3.Cursor:
        if kwargs.values():
            column_list = []
            values = []
            for key, value in kwargs.items():
                if value is None:
                    where = f"{key} IS NULL"
                else:
                    values.append(value)
                    where = f"{key} = ?"
                column_list.append(where)

            columns = " AND ".join(column_list)
            tuple_values = tuple(values)

            query = f"SELECT * FROM {self.table_name} WHERE {columns}"
        else:
            tuple_values = ()
            query = f"SELECT * FROM {self.table_name}"

        return self.cursor.execute(query, tuple_values)

    def insert(self, **kwargs) -> int:
        columns = ",".join(kwargs.keys())
        values = ",".join(["?" for _ in kwargs.keys()])

        query = f"INSERT INTO {self.table_name} ({columns}) "
        query += f"VALUES ({values})"
        tuple_values = tuple(kwargs.values())

        self.cursor.execute(query, tuple_values)
        self.connection.commit()
        return self.cursor.lastrowid

    def update(self, object_id=None, **kwargs) -> None:
        if not kwargs.values():
            return
        column_list = []
        for key in kwargs.keys():
            set_column = f"{key} = ?"
            column_list.append(set_column)

        columns = ", ".join(column_list)
        values = list(kwargs.values())

        query = f"UPDATE {self.table_name} SET {columns}"
        if object_id:
            query += f" WHERE id = ?"
            values.append(object_id)

        tuple_values = tuple(values)
        self.cursor.execute(query, tuple_values)
        self.connection.commit()

    def delete(self, object_id=None) -> None:
        query = f"DELETE FROM {self.table_name}"
        if object_id:
            query += " WHERE id = ?"
            tuple_values = (object_id,)
            self.cursor.execute(query, tuple_values)
        else:
            self.cursor.execute(query)

        self.connection.commit()



# First Party
from minislite.fields import DatabaseField
from minislite.models import MiniSLiteModel
from minislite.minislite import MiniSLiteDb

__all__ = ["MiniSLiteModel", "MiniSLiteDb", "DatabaseField"]



# Standard Library
from typing import List, Type

# First Party
from minislite.fields import DatabaseField
from minislite.models import MiniSLiteModel


class SqlManager:
    def __init__(self, model: Type[MiniSLiteModel]):
        self.model = model
        self.data_types = {str: "TEXT", int: "INTEGER", bool: "INTEGER", float: "REAL"}

    def generate_sql_script(self) -> str:
        table_name = self.model.get_table_name()
        field_lines = self.generate_fields_sql_lines()
        fields = ", ".join(field_lines)

        script = f'CREATE TABLE IF NOT EXISTS "{table_name}" ( {fields} );'
        return script

    def generate_fields_sql_lines(self) -> List[str]:
        field_list = self.model.get_fields()
        field_lines = ['"id" INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT']

        for field in field_list:
            field_name = field

            field_obj: DatabaseField = getattr(self.model, field)
            field_type = self.data_types.get(field_obj.field_type, None)
            if field_type is None:
                continue
            not_null = "NOT NULL" if field_obj.not_null else ""
            unique = "UNIQUE" if field_obj.unique else ""
            default = "DEFAULT {0}".format(field_obj.default) if field_obj.default is not None else ""
            auto_increment = "AUTOINCREMENT" if field_obj.auto_increment else ""

            line = f'"{field_name}" {field_type} {not_null} {unique} {default} {auto_increment}'
            field_lines.append(line)

        return field_lines



class DatabaseField:
    def __init__(self, field_type, default=None, unique=False, not_null=True, auto_increment=False) -> None:
        self.field_type = field_type
        self.default = default
        self.not_null = not_null
        self.unique = unique
        self.auto_increment = auto_increment



# Third Party
from setuptools import setup

setup(
    name="minislite",
    version="0.9",
    packages=["minislite"],
    url="https://github.com/ahmetkotan/minislite",
    license="",
    author="ahmetkotan",
    author_email="ahmtkotan@gmail.com",
    description="Mini SQLite ORM",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: SQL",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
    ],
)



# Standard Library
import os
import sqlite3
from typing import Type

# First Party
from minislite.sql import SqlManager
from minislite.models import MiniSLiteModel


class MiniSLiteDb:
    def __init__(self, db_path=":memory:"):
        self.db_path = db_path
        os.environ["MINILITE_DB_PATH"] = self.db_path

        self.connection = sqlite3.connect(self.db_path)
        self.connection.row_factory = sqlite3.Row

        self.cursor = self.connection.cursor()

    def add_model(self, model: Type[MiniSLiteModel]):
        sql_manager = SqlManager(model)
        sql_script = sql_manager.generate_sql_script()

        self.cursor.executescript(sql_script)
        self.connection.commit()

    def drop_model(self, model: Type[MiniSLiteModel]):
        table_name = model.get_table_name()
        self.cursor.execute(f"DROP TABLE {table_name}")
        self.connection.commit()

    def clean_tables(self):
        table_list = self.cursor.execute("SELECT name FROM sqlite_sequence").fetchall()
        for table in table_list:
            table_name = dict(table).get("name")
            self.cursor.execute(f"DELETE FROM {table_name}")

        self.connection.commit()
